<?php

namespace Drupal\graphql_compose\Form;

use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\graphql_compose\GraphQlFormatterPluginManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure the GraphQL formatters
 *
 * @internal
 */
class GraphQlComposeSettingsForm extends ConfigFormBase {

  /**
   * Config settings.
   *
   * @var string
   */
  const SETTINGS = 'graphql_compose.settings';

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity bundle info.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $entityTypeBundleInfo;

  /**
   * The entity bundle info.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $graphQlFormatterManager;

  /**
   * The current route match.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * List of availables GraphQL formatters
   *
   * @var array
   */
  protected $graphQlFormatters = [];

  /**
   * List of GraphQL default formatters.
   *
   * @var bool
   */
  protected $defaultFormatters = [
    'integer' => 'integer',
    'uuid' => 'uuid',
    'language' => 'string',
    'string' => 'string',
    'uri' => 'string',
    'list_integer' => 'list_integer',
    'timestamp' => 'string',
    'string_long' => 'string_long',
    'entity_reference' => '',
    'created' => 'created',
    'boolean' => 'boolean',
    'email' => 'email',
    'changed' => 'changed',
    'text_long' => 'text_long',
    'path' => 'path',
    'text_with_summary' => 'text_with_summary',
    'comment' => '',
    'image' => 'image',
    'link' => 'link',
    'file_uri' => '',
    'password' => 'password'
  ];

  /**
   * Constructs an \Drupal\views\Plugin\views\argument_validator\Entity object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity type bundle info.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityTypeBundleInfoInterface $entity_type_bundle_info, EntityFieldManagerInterface $entity_field_manager, GraphQlFormatterPluginManager $graphql_formatter_plugin_manager, RouteMatchInterface $route_match) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
    $this->entityFieldManager = $entity_field_manager;
    $this->graphQlFormatterManager = $graphql_formatter_plugin_manager;
    $this->routeMatch = $route_match;
    $this->graphQlFormatters = $graphql_formatter_plugin_manager->getDefinitions();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('entity_field.manager'),
      $container->get('plugin.manager.graphql_formatter'),
      $container->get('current_route_match')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'graphql_compose_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      static::SETTINGS . '.' . $this->routeMatch->getParameter('graphql_server')
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $graphql_server = NULL) {
    //dump($graphql_server);
    //$definition = \Drupal::service('config.typed')->getDefinition('graphql_compose.settings');
    $config = $this->config(static::SETTINGS . '.' . $graphql_server)->get('entity_types');
    $entity_types = $this->entityTypeManager->getDefinitions();
    $labels = [];
    $default = [];

    $bundles = $this->entityTypeBundleInfo->getAllBundleInfo();
    foreach ($entity_types as $entity_type_id => $entity_type) {
      if (!$entity_type instanceof ContentEntityTypeInterface || !isset($bundles[$entity_type_id])) {
        continue;
      }
      $labels[$entity_type_id] = $entity_type->getLabel() ?: $entity_type_id;
      $default[$entity_type_id] = $config[$entity_type_id]['exposed'] ? $entity_type_id : 0;
    }
    asort($labels);
    $form = [
      '#labels' => $labels,
      '#attached' => [
        'library' => [
          'graphql_compose/graphql_compose_admin'
        ],
      ],
      '#attributes' => [
        'class' => 'graphql-compose-settings-form',
      ],
    ];
    // Checkboxes of entity types
    $form['entity_types'] = [
      '#title' => $this->t('Available entities in GraphQL'),
      '#type' => 'checkboxes',
      '#options' => $labels,
      '#default_value' => $default,
    ];

    $form['settings'] = ['#tree' => TRUE];

    foreach ($labels as $entity_type_id => $label) {
      $entity_type = $entity_types[$entity_type_id];
      $form['settings'][$entity_type_id] = [
        '#title' => $label,
        '#type' => 'container',
        '#entity_type' => $entity_type_id,
        '#theme' => 'graphql_compose_settings_table',
        '#bundle_label' => $entity_type->getBundleLabel() ?: $label,
        '#states' => [
          'visible' => [
            ':input[name="entity_types[' . $entity_type_id . ']"]' => ['checked' => TRUE],
          ],
        ],
      ];
      // Build bundles form.
      foreach ($bundles[$entity_type_id] as $bundle => $bundle_info) {
        $form['settings'][$entity_type_id]['bundles'][$bundle] = [
          '#type' => 'item',
          '#label' => $bundle_info['label']
        ];
        $form['settings'][$entity_type_id]['bundles'][$bundle]['exposed'] = [
          '#type' => 'checkbox',
          '#default_value' => isset($config[$entity_type_id]['bundles'][$bundle]['exposed']) ?
            $config[$entity_type_id]['bundles'][$bundle]['exposed'] : 0,
        ];
        $form['settings'][$entity_type_id]['bundles'][$bundle]['prefix'] = [
          '#type' => 'textfield',
          '#title' => $this->t('GraphQL object name'),
          '#description' => $this->t('The GraphQL object name applied to the entity'),
          '#maxlength' => 10,
          '#size' => 30,
          '#default_value' => isset($config[$entity_type_id]['bundles'][$bundle]['prefix']) ?
            $config[$entity_type_id]['bundles'][$bundle]['prefix'] : $this->nameToCamelCase($entity_type_id . '_' . $bundle, TRUE),
          '#states' => [
            'visible' => [
              ':input[name="settings[' . $entity_type_id . '][bundles][' . $bundle . '][exposed]"]' => ['checked' => TRUE],
            ],
          ],
        ];
        $form['settings'][$entity_type_id]['bundles'][$bundle]['fields'] = [
          '#type' => 'container'
        ];
        $fields = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
        if ($fields) {
          // Build fields form.
          foreach ($fields as $field_name => $definition) {
            $types[] = $definition->getType();
            $form['settings'][$entity_type_id]['bundles'][$bundle]['fields'][$field_name]['exposed'] = [
              '#label' => $definition->getLabel(),
              '#type' => 'checkbox',
              '#default_value' => isset($config[$entity_type_id]['bundles'][$bundle]['fields'][$field_name]['exposed']) ?
                $config[$entity_type_id]['bundles'][$bundle]['fields'][$field_name]['exposed'] : 0,
            ];
            $form['settings'][$entity_type_id]['bundles'][$bundle]['fields'][$field_name]['formatter'] = [
              '#title' => $this->t('Formatter'),
              '#description' => $this->t('The GraphQL formatter applied to the field'),
              '#type' => 'select',
              '#options' => $this->getGrahQLFormattersOptions($definition->getType()),
              '#default_value' =>  isset($config[$entity_type_id]['bundles'][$bundle]['fields'][$field_name]['formatter']) ?
                $config[$entity_type_id]['bundles'][$bundle]['fields'][$field_name]['formatter'] : $this->getDefaultGrahQLFormatter($definition->getType()),
              '#states' => [
                'visible' => [
                  ':input[name="settings[' . $entity_type_id . '][bundles][' . $bundle . '][fields][' . $field_name . '][exposed]"]' => ['checked' => TRUE],
                ],
              ],
            ];
            $form['settings'][$entity_type_id]['bundles'][$bundle]['fields'][$field_name]['name'] = [
              '#title' => $this->t('Name'),
              '#description' => $this->t('The GraphQL field name'),
              '#type' => 'textfield',
              '#default_value' => isset($config[$entity_type_id]['bundles'][$bundle]['fields'][$field_name]['name']) ?
              $config[$entity_type_id]['bundles'][$bundle]['fields'][$field_name]['name'] : $this->nameToCamelCase($field_name),
              '#states' => [
                'visible' => [
                  ':input[name="settings[' . $entity_type_id . '][bundles][' . $bundle . '][fields][' . $field_name . '][exposed]"]' => ['checked' => TRUE],
                ],
              ],
            ];
          }

        }
      }
    }
    $form['actions']['#type'] = 'actions';
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save configuration'),
      '#button_type' => 'primary',
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $entity_type_settings = $form_state->getValue('entity_types');
    $settings = $form_state->getValue('settings');
    foreach ($settings as $entity_type => $entity_settings) {
      $settings[$entity_type]['exposed'] = $entity_type_settings[$entity_type] ? 1 : 0;
    }
    // Save the configuration.
    $this->config(static::SETTINGS . '.' . $this->routeMatch->getParameter('graphql_server'))
      ->set('entity_types', $settings)
      ->save();
    $this->messenger()->addStatus($this->t('Settings successfully updated.'));
  }

  /**
   * {@inheritdoc}
   */
  public function getGrahQLFormattersOptions($type) {
    $options = [];
    foreach ($this->graphQlFormatters as $id => $formatter) {
      if ($type == $formatter['type']) {
        $options[$id] = $id;
      }
    }
    return $options;
  }


  /**
   * {@inheritdoc}
   */
  public function getDefaultGrahQLFormatter($type) {
    return $this->defaultFormatters[$type];
  }

  /**
   * {@inheritdoc}
   */
  public function nameToCamelCase($string, $capitalizeFirstCharacter = false) {
    $str = str_replace('_', '', ucwords($string, '_'));
    if (!$capitalizeFirstCharacter) {
      $str = lcfirst($str);
    }
    return $str;
  }
}
